"""Copyright (c) 2023, Aydin Abdi.

This module is responsible for finding XML tags in XML files.
"""
from typing import List, Optional  # noqa F401
from xml.etree import ElementTree as ET

from qa_analytics_insights.xml_loader import XMLLoader


class XMLTagFinder:
    """Responsible for finding XML tags in XML files."""

    def __init__(self, xml_loader: XMLLoader) -> None:
        """Responsible for finding XML tags in XML files.

        Args:
            xml_loader: XMLLoader object.
        """
        self.xml_loader = xml_loader
        self.root = self.xml_loader.root  # type: ET.Element
        self._test_cases = []  # type: List[ET.Element]
        self._suite = None  # type: Optional[ET.Element]

    @property
    def test_cases(self) -> List[ET.Element]:
        """Returns all the testcase tags in the XML file.

        Returns:
            List of testcase tags.
        """
        if not self._test_cases:
            self._test_cases = self.root.findall(".//testcase")
        return self._test_cases

    @property
    def suite(self) -> ET.Element:
        """Returns the suite element.

        This method returns the testsuite element if it exists.
        If the testsuite element does not exist, it returns the root element.

        Returns:
            Suite element.
        """
        if self._suite is None:
            self._suite = self.root.find("testsuite") or self.root
        return self._suite



"""Copyright (c) 2023, Aydin Abdi.

This module is responsible for logging.
"""
import functools
import sys
import time
from typing import Any, Callable

from loguru import logger


def verbose_logging() -> None:
    """loguru verbose."""
    logger.remove()
    logger.add(sys.stderr, level="DEBUG")
    logger.add("outputs/qa_analytics_insights.log", level="DEBUG", rotation="1 day")


def default_logging() -> None:
    """loguru default."""
    logger.remove()
    logger.add(sys.stderr, level="INFO")
    logger.add("outputs/qa_analytics_insights.log", level="DEBUG", rotation="1 day")


def log_execution_time(method: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator to log the time execution of a method.

    Args:
        method: The method to be decorated.

    Returns:
        The decorated method.
    """

    @functools.wraps(method)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        """Wrapper function to log the time execution of a method.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            The result of the method.
        """
        start = time.time()
        result = method(*args, **kwargs)
        end = time.time()
        duration = end - start

        logger.info(f"{method.__name__} executed in {duration:.2f} seconds.")
        return result

    return wrapper



"""Copyright (c) 2023, Aydin Abdi.

This module is responsible for fetching paths from the given initial path.
"""
from pathlib import Path
from queue import Queue

from loguru import logger


class PathFetcher:
    """Responsible for fetching paths from the given initial path.

    Example:
        PathFetcher(initial_path='path/to/initial/path')
    """

    def __init__(self, initial_path: str) -> None:
        """Responsible for fetching paths from the given initial path.

        Args:
            initial_path: Initial path to fetch paths from.
        """
        self.initial_path = Path(initial_path)
        self.queue = Queue()  # type: Queue[Path]

    def fetch_paths(self) -> Queue[Path]:
        """Fetches paths from the given initial path.

        Returns:
            Queue of paths.
        """
        try:
            if self.initial_path.is_file():
                self.queue.put(self.initial_path)
            elif self.initial_path.is_dir():
                for file_path in self.initial_path.iterdir():
                    self.queue.put(file_path)
            else:
                logger.error(f"Invalid path: {self.initial_path}")
        except Exception as e:
            logger.exception(f"Error fetching paths: {str(e)}")
        return self.queue



"""Copyright (c) 2023, Aydin Abdi.

This module is analyze and extracts the data from
test results and can visualize the data in different ways.
"""
from qa_analytics_insights._version import version

__author__ = "Aydin Abdi"
__license__ = "MIT"
__version__ = version



"""Copyright (c) 2023, Aydin Abdi.

This module is responsible for parsing test cases from XML tags.
"""
import re
from typing import Optional
from xml.etree import ElementTree as ET

from loguru import logger

from qa_analytics_insights.data_classes import TestCase, TestSuite


class ParserTestSuite:
    """Responsible for parsing TestSuite objects from XML tags."""

    def __init__(self, test_suite: ET.Element) -> None:
        """Responsible for parsing TestSuite objects from XML tags.

        Args:
            test_suite: XML tag of the test suite.
        """
        self.test_suite = test_suite

    def parse(self) -> TestSuite:
        """Parses the XML tag into TestSuite object.

        Returns:
            TestSuite object.
        """
        suite_name = self.test_suite.attrib.get("name", "")
        suite_tests = int(self.test_suite.attrib.get("tests", 0))
        suite_errors = int(self.test_suite.attrib.get("errors", 0))
        suite_failures = int(self.test_suite.attrib.get("failures", 0))
        suite_skipped = int(
            self.test_suite.attrib.get("skip")
            or self.test_suite.attrib.get("skipped")
            or 0
        )
        suite_execution_time = float(self.test_suite.attrib.get("time", 0))
        suite_timestamp = self.test_suite.attrib.get("timestamp", "")

        test_suite = TestSuite(
            name=suite_name,
            tests=suite_tests,
            errors=suite_errors,
            failures=suite_failures,
            skipped=suite_skipped,
            execution_time=suite_execution_time,
            timestamp=suite_timestamp,
        )

        return test_suite


class ParserTestCase:
    """Responsible for parsing TestCase objects from XML tags."""

    def __init__(self, test_case: ET.Element) -> None:
        """Responsible for parsing TestCase objects from XML tag.

        Args:
            test_case: XML tag of the test case.
        """
        self.test_case = test_case

    def parse(self) -> TestCase:
        """Parses the XML tag into TestCase object.

        Returns:
            TestCase object.
        """
        test_class_name = self.test_case.attrib.get("classname", "")
        if test_class_name:
            test_class_name = test_class_name.split(".")[-1]

        test_module_name = self.test_case.attrib.get("classname", "")
        if test_module_name:
            test_module_name = test_module_name.split(".")[-2]

        test_case_name = self.test_case.attrib.get("name", "")
        test_case_time = float(self.test_case.attrib.get("time", 0))
        test_case_result = self.get_testcase_result()

        failure_reason = None
        error_reason = None
        skipped_reason = None
        if test_case_result == "failed":
            failure_reason = self.get_failure_reason()
        if test_case_result == "error":
            error_reason = self.get_failure_reason(tag="error")

        if test_case_result == "skipped":
            skipped_reason = self.get_skipped_reason()

        system_out = self.get_system_out()
        time_stamp = self.get_timestamp()

        return TestCase(
            name=test_case_name,
            test_module=test_module_name,
            test_class=test_class_name,
            execution_time=test_case_time,
            result=test_case_result,
            timestamp=time_stamp,
            failure_reason=failure_reason,
            error_reason=error_reason,
            skipped_reason=skipped_reason,
            system_out=system_out,
        )

    def get_testcase_result(self) -> str:
        """Parses the test case result from the test case tag.

        Returns:
            Test case result.
        """
        result = "passed"
        if self.test_case.find("failure") is not None:
            result = "failed"
        elif self.test_case.find("skipped") is not None:
            result = "skipped"
        elif self.test_case.find("error") is not None:
            result = "error"
        return result

    def find_tag_attribute(
        self, tag: str, attrib: Optional[str] = None
    ) -> Optional[str]:
        """Find the tag for test case element and return the attribute or text.

        Args:
            tag: Tag to parse.
            attrib: Attribute to parse. If None, the tag text is returned.

        Returns:
            Find the tag and return the attribute or text.
        """
        found_tag = self.test_case.find(tag)
        if found_tag is None:
            return None
        if attrib is not None:
            attrib_text = found_tag.get(attrib)
            return attrib_text
        return found_tag.text

    def get_failure_reason(self, tag: str = "failure") -> Optional[str]:
        """Parse the failure reason from a failure tag.

        Args:
            tag: XML element representing a failure tag. Defaults to "failure".

        Returns:
            Failure reason.
        """
        failure_message = self.find_tag_attribute(tag, "message")
        if not failure_message:
            logger.debug(
                f"{tag.upper()} message not found in {tag} tag. "
                f"Test case: {self.test_case.text}"
            )
            return None
        try:
            # Try to extract the failure reason from the failure message
            split_by_captured_logging = failure_message.strip().split("\n")
            if split_by_captured_logging:
                failure_reason = split_by_captured_logging[0]
                return failure_reason
        except Exception as message_parse_exception:
            logger.warning(
                f"Could not parse {tag} reason from {tag} message."
                f"Exception: {message_parse_exception}"
            )
            logger.warning(f"Test case: {self.test_case.text}")
        return failure_message

    def get_skipped_reason(self, tag: str = "skipped") -> Optional[str]:
        """Parse the skipped reason from a skipped tag.

        Args:
            tag: XML element representing a skipped tag. Defaults to "skipped".

        Returns:
            Skipped reason.
        """
        skipped_message = self.find_tag_attribute(tag, "message")
        if skipped_message:
            return skipped_message

        logger.debug(
            "Skipped message not found in skipped tag. Test case: {test_case.text}"
        )
        return None

    def get_timestamp(self, tag: str = "system-out") -> Optional[str]:
        """Parse the timestamp from a system-out tag or timestamp tag.

        Args:
            tag: XML element representing a system-out tag. Defaults to "system-out".

        Returns:
            Timestamp or None if no timestamp is found.
        """
        # check if any timestamp tag exists
        timestamp_found = self.find_tag_attribute(tag="timestamp")
        if timestamp_found is not None:
            logger.debug(f"Tag timestamp found: {timestamp_found}")
            return timestamp_found
        logger.debug(f"Tag timestamp not found: {timestamp_found}")
        logger.debug(f"Parsing timestamp from {tag} tag...")

        system_out_tag = self.find_tag_attribute(tag)
        # Sometimes the timestamp is in the first line of the system-out tag
        if system_out_tag:
            timestamp_text = system_out_tag.strip().split("\n")[0]
            # Extract timestamp using regex
            # assuming it's in the format YYYYMMDD HH:MM:SS
            timestamp_match = re.match(r"(\d{8} \d{2}:\d{2}:\d{2})", timestamp_text)
            if timestamp_match:
                logger.debug(f"Found timestamp: {timestamp_match.group(0)}")
                return timestamp_match.group(0)

        logger.debug(f"Timestamp not parsed. Unexpected system-out: {system_out_tag}.")
        return None

    def get_system_out(self, tag: str = "system-out") -> Optional[str]:
        """Parse the system out from a system-out tag.

        Args:
            tag: Tag to parse.

        Returns:
            System out or None if no system out is found.
        """
        system_out_tag = self.find_tag_attribute(tag)

        # The system out is the text after the CDATA
        if system_out_tag:
            system_out_text = system_out_tag
            return system_out_text

        logger.debug(
            f"Could not parse system out from system-out tag: {system_out_tag}"
        )
        return None



"""Sphinx configuration."""
import inspect
import os
import sys
import shutil

from docutils import nodes

from qa_analytics_insights._version import __version__


def update_version_badge(app: object, doctree: object, docname: str) -> None:
    """Update the version badge in the documentation to match the package version.

    This function is called by Sphinx when the doctree-read event is emitted.
    This event is emitted when the doctree is read from disk and ready to be
    processed. See: https://www.sphinx-doc.org/en/master/extdev/appapi.html#sphinx-core-events
    
    Args:
        app: The Sphinx application object.
        doctree: The document tree.
        docname: The document name.
    """
    # Extract the version information from the package
    package_version = __version__
    
    for image_node in doctree.traverse(nodes.image):
        original_url = image_node['uri']

        # Identify the image node to update by matching a unique part of its URL
        if 'img.shields.io/badge/qa_analytics_insights-<version>-blue' in original_url:
            # Create a new URL with the updated version information
            new_url = (
                f"https://img.shields.io/badge/qa_analytics_insights-{package_version}-blue"
            )
            
            # Update the image node URL
            image_node['uri'] = new_url

def setup(app):
    app.connect('doctree-resolved', update_version_badge)

# -- Path setup --------------------------------------------------------------

__location__ = os.path.join(os.getcwd(), os.path.dirname(inspect.getfile(inspect.currentframe())))

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.join(__location__, "../src"))

# -- Run sphinx-apidoc -------------------------------------------------------
# This hack is necessary since RTD does not issue `sphinx-apidoc` before running
# `sphinx-build -b html . _build/html`. See Issue:
# https://github.com/readthedocs/readthedocs.org/issues/1139
# DON'T FORGET: Check the box "Install your project inside a virtualenv using
# setup.py install" in the RTD Advanced Settings.
# Additionally it helps us to avoid running apidoc manually

try:  # for Sphinx >= 1.7
    from sphinx.ext import apidoc
except ImportError:
    from sphinx import apidoc

output_dir = os.path.join(__location__, "api")
module_dir = os.path.join(__location__, "../src/qa_analytics_insights")
try:
    shutil.rmtree(output_dir)
except FileNotFoundError:
    pass

try:
    import sphinx

    cmd_line = f"sphinx-apidoc --implicit-namespaces -f -o {output_dir} {module_dir}"

    args = cmd_line.split(" ")
    if tuple(sphinx.__version__.split(".")) >= ("3", "5", "2"):
        # This is a rudimentary parse_version to avoid external dependencies
        args = args[1:]

    apidoc.main(args)
except Exception as e:
    print(f"Running `sphinx-apidoc` failed!\n{e}")

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.todo",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.ifconfig",
    "sphinx.ext.mathjax",
    "sphinx_rtd_theme",
    "sphinx.ext.napoleon",
    "sphinx_tabs.tabs",
    "sphinx-prompt",
    "sphinx_toolbox",
]

github_username = "aydabd"
github_repository = "qa-analytics-insights"
autodoc_mock_imports = ["loguru"]
# autodoc_typehints = "none"
always_document_param_types = True
typehints_defaults = "braces-after"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "qa-analytics-insights"
copyright = "2023, Aydin Abdi"
author = "Aydin Abdi"

# The full version, including alpha/beta/rc tags.
version = __version__
if not version or version.lower() == "unknown":
    version = os.getenv("READTHEDOCS_VERSION", "unknown")  # automatically set by RTD

release = version


# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", ".venv"]

# If this is True, todo emits a warning for each TODO entries. The default is False.
todo_emit_warnings = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "canonical_url": "",
    "analytics_id": "UA-XXXXXXX-1",  #  Provided by Google in your dashboard
    "logo_only": False,
    "display_version": True,
    "prev_next_buttons_location": "bottom",
    "style_external_links": False,
    "style_nav_header_background": "#2980B9",
    # Toc options
    "collapse_navigation": True,
    "sticky_navigation": True,
    "navigation_depth": 2,
    "includehidden": True,
    "titles_only": False,
}

# Output file base name for HTML help builder.
htmlhelp_basename = "qa-analytics-insights-doc"

print(f"loading configurations for {project} {version} ...", file=sys.stderr)




"""Copyright (c) 2023, Aydin Abdi.

This module is responsible for loading XML files.
"""
from typing import Optional  # noqa F401
from xml.etree import ElementTree as ET


class XMLLoader:
    """Responsible for loading XML file."""

    def __init__(self, xml_path: str) -> None:
        """Responsible for loading XML files into XMLParser.

        Args:
            xml_path: Path to the XML file.
        """
        self.xml_path = xml_path
        self._tree = None  # type: Optional[ET.ElementTree]
        self._root = None  # type: Optional[ET.Element]

    @property
    def tree(self) -> ET.ElementTree:  # pragma: no cover
        """Returns the XML tree.

        Returns:
            XML tree.
        """
        if not self._tree:
            self._tree = ET.parse(self.xml_path)
        return self._tree

    @property
    def root(self) -> ET.Element:  # pragma: no cover
        """Returns the XML root.

        Returns:
            XML root.
        """
        if not self._root:
            self._root = self.tree.getroot()
        return self._root



"""Copyright (c) 2023, Aydin Abdi.

Unit-tests for the xml_parser module.
"""
from qa_analytics_insights.xml_loader import XMLLoader
from qa_analytics_insights.xml_parser import XMLParser
from qa_analytics_insights.xml_tag_finder import XMLTagFinder

nosetests_xml = "tests/data/nosetests_test_result.xml"
pytests_xml = "tests/data/pytest_test_result.xml"
invalid_file = "tests/data/text.txt"


def test_parse_with_pytest_xml():
    """Tests the `parse` method for pytest XML files."""
    xml_loader = XMLLoader(pytests_xml)
    xml_tag_finder = XMLTagFinder(xml_loader)
    parser = XMLParser(xml_tag_finder)
    test_suite = parser.parse()

    assert test_suite.name == "qa-analytics-insights"
    assert test_suite.tests == 3
    assert test_suite.errors == 0
    assert test_suite.failures == 1
    assert test_suite.skipped == 1
    assert len(test_suite.test_classes) == 2
    assert len(test_suite.test_cases) == 0
    assert test_suite.passed == 1


def test_parse_with_nosetests_xml():
    """Tests the `parse` method for nosetests XML files."""
    xml_loader = XMLLoader(nosetests_xml)
    xml_tag_finder = XMLTagFinder(xml_loader)
    parser = XMLParser(xml_tag_finder)
    test_suite = parser.parse()

    assert test_suite.name == "nosetests"
    assert test_suite.tests == 6
    assert test_suite.errors == 1
    assert test_suite.failures == 1
    assert test_suite.skipped == 1
    assert len(test_suite.test_classes) == 5
    assert len(test_suite.test_cases) == 1
    assert test_suite.passed == 3



"""Copyright (c) 2023, Aydin Abdi.

Unit tests for log.py module.
"""
import time

from qa_analytics_insights.log import (
    default_logging,
    log_execution_time,
    verbose_logging,
)


def test_default_logging() -> None:
    """Test default_logging function."""
    assert default_logging() is None


def test_verbose_logging() -> None:
    """Test verbose_logging function."""
    assert verbose_logging() is None


def test_log_execution_time() -> None:
    """Test log_execution_time decorator."""

    @log_execution_time
    def execution_time_func() -> None:
        """Test function."""
        time.sleep(1)

    execution_time_func()



"""Copyright (c) 2023, Aydin Abdi.

This module is responsible for visualizing the results of the parsed TestClass objects.
"""
from __future__ import annotations

import base64
import math
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from io import BytesIO
from pathlib import Path
from typing import List, Optional

import matplotlib
import numpy as np
from loguru import logger
from matplotlib import pyplot as plt

from qa_analytics_insights.data_classes import TestCase, TestClass, TestSuite

# Don't use GUI backend for matplotlib to avoid errors when running on a server.
matplotlib.use('Agg')


class ResultVisualizer:
    """Class for visualizing test results."""

    def __init__(
        self,
        test_suites: Optional[List[TestSuite]] = None,
    ) -> None:
        """Initialize the TestResultVisualizer class.

        Args:
            test_suites: List of TestSuite objects.
        """
        self.test_suites = test_suites
        self._test_classes = []  # type: List[TestClass]
        self._test_cases = []  # type: List[TestCase]
        self.plot = plt

    @property
    def test_classes(self) -> List[TestClass]:
        """Return the test classes.

        Returns:
            List of TestClass objects.
        """
        if not self._test_classes:
            if not self.test_suites or self.test_suites is None:
                logger.warning("No test suites found.")
                return []
            for test_suite in self.test_suites:
                self._test_classes.extend(test_suite.test_classes)
        return self._test_classes

    @property
    def test_cases(self) -> List[TestCase]:
        """Return the test cases.

        Returns:
            List of TestCase objects.
        """
        if not self._test_cases:
            if not self.test_suites or self.test_suites is None:
                logger.warning("No test suites found.")
                return []
            for test_suite in self.test_suites:
                self._test_cases.extend(test_suite.test_cases)
            for test_class in self.test_classes:
                self._test_cases.extend(test_class.test_cases)
        return self._test_cases

    def plot_pie_charts_test_classes(self) -> Optional[plt.Figure]:
        """Plot pie charts for each test class.

        Returns:
            Figure containing the pie charts.
        """
        logger.info("Plotting pie charts for test results.")
        if not self.test_classes:
            logger.warning("No test classes found.")
            return None

        num_test_classes = len(self.test_classes)
        num_rows = math.ceil(math.sqrt(num_test_classes))
        num_cols = math.ceil(num_test_classes / num_rows)

        # Set a minimum size for the pie chart
        min_size_x, min_size_y = 3, 3  # you can adjust these as needed
        pix_size_x = max(min_size_x, num_test_classes * 2)
        pix_size_y = max(min_size_y, num_test_classes / 2)

        pie_charts, axs = self.plot.subplots(
            num_rows, num_cols, figsize=(pix_size_x, pix_size_y), clear=True
        )

        # Check if axs is an array and flatten it
        if isinstance(axs, np.ndarray):
            axs = axs.flatten()
        elif isinstance(axs, plt.Axes):
            axs = np.array([axs])

        for ax, test_class in zip(axs, self.test_classes):
            statuses = {"skipped": 0, "failed": 0, "passed": 0, "error": 0}
            colors = {
                "skipped": "gray",
                "failed": "red",
                "passed": "green",
                "error": "orange",
            }
            for test_case in test_class.test_cases:
                statuses[test_case.result] += 1
            labels = list(statuses.keys())
            sizes = [statuses[key] for key in labels]
            labels_with_counts = [
                f"{label} ({count})" for label, count in zip(labels, sizes)
            ]
            ax.pie(
                sizes,
                autopct='%1.1f%%',
                startangle=180,
                colors=[colors[key] for key in labels],
                textprops={'fontsize': 6},
            )
            ax.axis('equal')
            ax.set_title(f"{test_class.name}")
            ax.legend(labels_with_counts, loc='upper right', fontsize=6)

        # Remove unused subplots
        for i in range(num_test_classes, num_rows * num_cols):
            pie_charts.delaxes(axs[i])

        return pie_charts

    def plot_failed_test_cases_table(self) -> Optional[plt.Figure]:
        """Plot a table of failed/error test cases.

        returns:
            The figure for failed/error test cases table.
        """
        logger.info("Creating failed/error test cases table.")
        all_failed_tests = []

        if self.test_cases:
            for testcase in self.test_cases:
                if testcase.result == "failed" or testcase.result == "error":
                    all_failed_tests.append(
                        (
                            testcase.test_class or "N/A",
                            testcase.name,
                            testcase.failure_reason or testcase.error_reason,
                        )
                    )

        if not all_failed_tests:
            logger.warning("No failed test cases found.")
            return None

        lenght_failed_tests = len(all_failed_tests)
        pixels_per_row = lenght_failed_tests * 2
        pixels_per_column = lenght_failed_tests / 4

        failed_test_cases_table, axes = self.plot.subplots(
            figsize=(pixels_per_row, pixels_per_column)
        )
        axes.axis('tight')
        axes.axis('off')
        columns = ("Test Class", "Test Name", "Failure Reason")
        table = axes.table(
            cellText=all_failed_tests,
            colLabels=columns,
            cellLoc='left',
            loc='left',
            colColours=['#DDEBF7'] * len(columns),
        )
        table.auto_set_font_size(False)
        table.set_fontsize(8)

        # Loop over cells to adjust the cell borders and wrap text for better formatting
        for _, cell in table.get_celld().items():
            cell.set_edgecolor("black")
            cell.set_linewidth(0.5)
            text = cell.get_text().get_text()
            cell.set_text_props(wrap=True)
            cell._text.set_text(text)

        # Adjust layout to make sure everything fits
        table.auto_set_column_width(col=list(range(len(columns))))
        table.scale(1.5, 1.5)  # Increase the size of cells to fit text

        return failed_test_cases_table

    def plot_skipped_test_cases_table(self) -> Optional[plt.Figure]:
        """Plot a table of skipped test cases.

        returns:
            The figure for skipped test cases table.
        """
        logger.info("Creating skipped test cases table.")
        all_skipped_tests = []

        if self.test_cases:
            for test_case in self.test_cases:
                if test_case.result == "skipped":
                    all_skipped_tests.append(
                        (
                            test_case.test_class or "N/A",
                            test_case.name,
                            test_case.skipped_reason,
                        )
                    )

        if not all_skipped_tests:
            logger.warning("No skipped test cases found.")
            return None

        lenght_skipped_tests = len(all_skipped_tests)
        pixels_per_row = lenght_skipped_tests * 2
        pixels_per_column = lenght_skipped_tests / 4

        skipped_test_cases_table, axes = self.plot.subplots(
            figsize=(pixels_per_row, pixels_per_column)
        )
        axes.axis('tight')
        axes.axis('off')
        columns = ("Test Class", "Test Name", "Skipped Reason")
        table = axes.table(
            cellText=all_skipped_tests,
            colLabels=columns,
            cellLoc='left',
            loc='left',
            colColours=['#DDEBF7'] * len(columns),
        )
        table.auto_set_font_size(False)
        table.set_fontsize(8)

        # Loop over cells to adjust the cell borders and wrap text for better formatting
        for _, cell in table.get_celld().items():
            cell.set_edgecolor("black")
            cell.set_linewidth(0.5)
            text = cell.get_text().get_text()
            cell.set_text_props(wrap=True)
            cell._text.set_text(text)

        # Adjust layout to make sure everything fits
        table.auto_set_column_width(col=list(range(len(columns))))
        table.scale(1.5, 1.5)  # Increase the size of cells to fit text

        return skipped_test_cases_table

    def plot_top_slowest_test_classes_pie_bar_chart(
        self, slowest_test_classes: Optional[List[TestClass]] = None
    ) -> Optional[plt.Figure]:
        """Plot a pie bar chart of the top slowest test classes.

        Args:
            slowest_test_classes: List of the slowest test classes.

        Returns:
            The figure for the top slowest test classes pie bar chart.
        """
        logger.info("Creating top slowest test classes pie bar chart.")
        if not slowest_test_classes or slowest_test_classes is None:
            logger.warning("No slowest test classes found.")
            return None

        labels = [test_class.name for test_class in slowest_test_classes]
        sizes = [test_class.execution_time for test_class in slowest_test_classes]

        min_height = 5  # Set a minimum figure height
        fig_height = max(min_height, len(labels))

        top_slowest_test_classes_pie_bar_chart, axes = self.plot.subplots(
            figsize=(20, fig_height), clear=True
        )

        # You can add variety to the color if you want
        colors = ['green' for _ in labels]

        axes.bar(labels, sizes, color=colors)
        axes.set_ylabel("Execution Time (seconds)", fontsize=12)
        axes.set_xlabel("Test Class", fontsize=12)

        return top_slowest_test_classes_pie_bar_chart

    @staticmethod
    def figure_to_base64(figure: plt.Figure) -> str:
        """Convert a Matplotlib figure to a base64 string.

        Args:
            figure: The figure to convert.

        Returns:
            The base64 encoded string representation of the figure.
        """
        img = BytesIO()
        figure.savefig(img, format="svg", bbox_inches="tight")
        img.seek(0)
        return base64.b64encode(img.read()).decode()

    def build_subplots_pie_chart_test_classes_results(self) -> Optional[str]:
        """Plot the results of the parsed TestClass objects.

        Returns:
            The base64 encoded string representation of the figure.
        """
        # Create the pie chart figure
        pie_charts = self.plot_pie_charts_test_classes()
        if pie_charts is None:
            logger.debug("No test classes found.")
            return None
        return self.figure_to_base64(pie_charts)

    def build_subplots_failed_table(self) -> Optional[str]:
        """Plot a table of failed test cases.

        Returns:
            The base64 encoded string representation of the figure.
        """
        failed_test_cases_table = self.plot_failed_test_cases_table()
        if failed_test_cases_table is None:
            logger.debug("No failed test cases found.")
            return None
        return self.figure_to_base64(failed_test_cases_table)

    def build_subplots_skipped_table(self) -> Optional[str]:
        """Plot a table of skipped test cases.

        Returns:
            The base64 encoded string representation of the figure.
        """
        skipped_test_cases_table = self.plot_skipped_test_cases_table()
        if skipped_test_cases_table is None:
            logger.debug("No skipped test cases found.")
            return None
        return self.figure_to_base64(skipped_test_cases_table)

    def build_subplots_top_slowest_test_classes(
        self, slowest_test_classes: List[TestClass]
    ) -> Optional[str]:
        """Plot a pie bar chart of the top slowest test classes.

        Args:
            slowest_test_classes: List of the slowest test classes.

        Returns:
            The base64 encoded string representation of the figure.
        """
        slowest_tests_bar_chart = self.plot_top_slowest_test_classes_pie_bar_chart(
            slowest_test_classes
        )
        if slowest_tests_bar_chart is None:
            logger.debug("No slowest test classes found.")
            return None
        return self.figure_to_base64(slowest_tests_bar_chart)

    def plot_test_suites_summary_table(self) -> Optional[plt.Figure]:
        """Plot a table of test suites summary.

        Returns:
            The figure for test suites summary table.
        """
        logger.info("Creating test suites summary table.")
        test_suites_summary = []
        if self.test_suites:
            for test_suite in self.test_suites:
                test_suites_summary.append(
                    (
                        test_suite.name,
                        test_suite.tests,
                        test_suite.passed,
                        test_suite.failures,
                        test_suite.skipped,
                        test_suite.errors,
                        test_suite.execution_time,
                    )
                )
            test_suites_summary.append(
                (
                    "Total",
                    sum([test_suite.tests for test_suite in self.test_suites]),
                    sum([test_suite.passed for test_suite in self.test_suites]),
                    sum([test_suite.failures for test_suite in self.test_suites]),
                    sum([test_suite.skipped for test_suite in self.test_suites]),
                    sum([test_suite.errors for test_suite in self.test_suites]),
                    sum([test_suite.execution_time for test_suite in self.test_suites]),
                )
            )
            failure_rate = round(
                sum(
                    [
                        test_suite.failures + test_suite.errors
                        for test_suite in self.test_suites
                    ]
                )
                / sum([test_suite.tests for test_suite in self.test_suites])
                * 100,
                2,
            )
            test_suites_summary.append(
                (  # type: ignore
                    "Failure rate",
                    f"{failure_rate}%",
                    "",
                    "",
                    "",
                    "",
                    "",
                )
            )

        if not test_suites_summary:
            logger.warning("No test suites found.")
            return None

        lenght_test_suites_summary = len(test_suites_summary)

        # Set a minimum size for the bar chart
        min_size_x, min_size_y = 1, 1
        pixels_per_row = max(min_size_x, lenght_test_suites_summary * 2)
        pixels_per_column = max(min_size_y, lenght_test_suites_summary / 2)

        test_suites_summary_table, axes = self.plot.subplots(
            figsize=(pixels_per_row, pixels_per_column)
        )
        axes.axis('tight')
        axes.axis('off')
        columns = (
            "Test Suite",
            "Total",
            "Passed",
            "Failed",
            "Skipped",
            "Errors",
            "Time",
        )
        table = axes.table(
            cellText=test_suites_summary,
            colLabels=columns,
            cellLoc='left',
            loc='left',
            colColours=['#DDEBF7'] * len(columns),
        )
        table.auto_set_font_size(False)
        table.set_fontsize(8)

        # Loop over cells to adjust the cell borders and wrap text for better formatting
        for _, cell in table.get_celld().items():
            cell.set_edgecolor("black")
            cell.set_linewidth(0.5)
            text = cell.get_text().get_text()
            cell.set_text_props(wrap=True)
            cell._text.set_text(text)

        # Adjust layout to make sure everything fits
        table.auto_set_column_width(col=list(range(len(columns))))
        table.scale(1.5, 1.5)

        return test_suites_summary_table

    def build_subplots_test_suites_summary_table(self) -> Optional[str]:
        """Plot a table of test suites summary.

        Returns:
            The base64 encoded string representation of the figure.
        """
        test_suites_summary_table = self.plot_test_suites_summary_table()
        if test_suites_summary_table is None:
            logger.debug("No test suites found.")
            return None
        return self.figure_to_base64(test_suites_summary_table)

    @staticmethod
    def generate_html_report_to_file(
        output: str,
        pie_charts: Optional[str] = None,
        failed_table: Optional[str] = None,
        skipped_table: Optional[str] = None,
        summary_table: Optional[str] = None,
        slowest_classes: Optional[str] = None,
    ) -> None:
        """Generate an HTML report of the test results.

        Args:
            pie_charts: Base64 string representing the pie charts.
            failed_table: Base64 string representing the failed test cases table.
            skipped_table: Base64 string representing the skipped test cases table.
            summary_table: Base64 string representing the test suites summary table.
            slowest_classes: Base64 string representing the top slowest test classes.
            output: Output file name.
        """
        logger.info("Generating HTML report...")
        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        html_content = f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>Test Results Visualization</title>
        </head>
        <body>
            <h1>Generated Test Result({timestamp})</h1>
            <h2>Test Suites Summary</h2>
            <img
                src="data:image/svg+xml;base64,{summary_table}"
                alt="Test Suites Summary Table">
            <h2>Test Results Pie Charts Based on Test Classes</h2>
            <img
                src="data:image/svg+xml;base64,{pie_charts}"
                alt="Test Results Pie Charts">
            <h2>Failed Test Cases</h2>
            <img
                src="data:image/svg+xml;base64,{failed_table}"
                alt="Failed Test Cases Table">
            <h2>Top Slowest Test Classes</h2>
            <img
                src="data:image/svg+xml;base64,{slowest_classes}"
                alt="Top Slowest Test Classes Pie Bar Chart">
            <h2>Skipped Test Cases</h2>
            <img src="data:image/svg+xml;base64,{skipped_table}"
                alt="Skipped Test Cases Table">
        </body>
        </html>
        """

        # Save the HTML content to a file
        html_file_path = f"{output}.html"
        with open(html_file_path, "w") as f:
            f.write(html_content)
        logger.info(f"HTML report saved to: {Path(html_file_path).absolute()}")

    def run(
        self,
        output: str = "test_results_visualization",
        pie_charts: Optional[plt.Figure] = None,
        failed_table: Optional[plt.Figure] = None,
        skipped_table: Optional[plt.Figure] = None,
        summary_table: Optional[plt.Figure] = None,
        slowest_classes: Optional[List[TestClass]] = None,
    ) -> None:
        """Main execution method.

        Args:
            output: Output file name.
            pie_charts: Figure containing the pie charts.
            failed_table: Figure containing the failed test cases table.
            skipped_table: Figure containing the skipped test cases table.
            slowest_classes: List of the slowest test classes.
            summary_table: Figure containing the test suites summary table.
        """
        if pie_charts:
            pie_charts = self.build_subplots_pie_chart_test_classes_results()
        if failed_table:
            failed_table = self.build_subplots_failed_table()
        if skipped_table:
            skipped_table = self.build_subplots_skipped_table()
        if summary_table:
            summary_table = self.build_subplots_test_suites_summary_table()
        if slowest_classes:
            slowest_test_classes = self.build_subplots_top_slowest_test_classes(
                slowest_classes
            )

        self.generate_html_report_to_file(
            output,
            pie_charts,
            failed_table,
            skipped_table,
            summary_table,
            slowest_test_classes,
        )


class ParallelResultVisualizer(ResultVisualizer):
    """Class to visualize the results of the parsed TestClass objects in parallel."""

    def __init__(
        self,
        test_suites: Optional[List[TestSuite]] = None,
    ) -> None:
        """Initialize the ParallelTestResultVisualizer class.

        Args:
            test_classes: List of TestClass objects.
        """
        super().__init__(test_suites)

    def generate_html_plots(
        self,
        output: str,
        slowest_test_classes: Optional[List[TestClass]] = None,
    ) -> None:
        """Main execution method.

        Args:
            output: Output file name.
            slowest_test_classes: List of the slowest test classes.
        """
        slowest_classes_base64 = None
        # Use ThreadPoolExecutor to run plot generation in parallel
        with ThreadPoolExecutor(max_workers=4) as executor:
            pie_chart_future = executor.submit(
                self.build_subplots_pie_chart_test_classes_results
            )
            failed_table_future = executor.submit(self.build_subplots_failed_table)
            skipped_table_future = executor.submit(self.build_subplots_skipped_table)
            summary_table_future = executor.submit(
                self.build_subplots_test_suites_summary_table
            )
            if slowest_test_classes and slowest_test_classes is not None:
                slowest_classes_future = executor.submit(
                    self.build_subplots_top_slowest_test_classes, slowest_test_classes
                )
                slowest_classes_base64 = slowest_classes_future.result()

            # Wait for all futures to complete and retrieve results
            pie_chart_base64 = pie_chart_future.result()
            failed_table_base64 = failed_table_future.result()
            skipped_table_base64 = skipped_table_future.result()
            summary_table_base64 = summary_table_future.result()

        # Generate the HTML report with the results
        self.generate_html_report_to_file(
            output,
            pie_chart_base64,
            failed_table_base64,
            skipped_table_base64,
            summary_table_base64,
            slowest_classes_base64,
        )



"""Copyright (c) 2023, Aydin Abdi.

This file contains the data classes used in the qa-analytics-insights package.
"""
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from typing import List, Optional


@dataclass
class TestCase:
    """Represents a test case.

    Args:
        name: Name of the test case.
        test_module: Name of the test module.
        test_class: Name of the test class.
        execution_time: Execution time of the test case.
        result: Result of the test case.
        timestamp: Timestamp of the test case.
        failure_reason: Failure reason of the test case.
        skipped_reason: Skipped reason of the test case.
        system_out: System out of the test case.
    """

    name: str
    test_module: Optional[str] = None
    test_class: Optional[str] = None
    execution_time: float = 0.0
    result: str = "passed"
    timestamp: Optional[str] = None
    failure_reason: Optional[str] = None
    error_reason: Optional[str] = None
    skipped_reason: Optional[str] = None
    system_out: Optional[str] = None


@dataclass
class TestClass:
    """Represents a test class.

    Args:
        name: Name of the test class.
        test_cases: List of test cases in the test class.
        passed: Number of passed tests in the test class.
        failed: Number of failed tests in the test class.
        skipped: Number of skipped tests in the test class.
        errors: Number of errors in the test class.
        execution_time: Execution time of the test class.
        failed_test_cases: List of failed test cases in the test class.
        skipped_test_cases: List of skipped test cases in the test class.
        error_test_cases: List of error test cases in the test class.
    """

    name: str
    test_cases: List[TestCase] = field(default_factory=list)
    passed: int = field(init=False, default=0)
    failed: int = field(init=False, default=0)
    skipped: int = field(init=False, default=0)
    errors: int = field(init=False, default=0)
    execution_time: float = field(init=False, default=0.0)
    failed_test_cases: List[TestCase] = field(init=False, default_factory=list)
    skipped_test_cases: List[TestCase] = field(init=False, default_factory=list)
    error_test_cases: List[TestCase] = field(init=False, default_factory=list)

    def __post_init__(self) -> None:
        """Post initialization method.

        This method calculates the number of passed, failed, skipped and
        error tests in the test class.
        """
        with ThreadPoolExecutor() as executor:
            executor.map(self._calculate_test_class_metrics, self.test_cases)

    def _calculate_test_class_metrics(self, test_case: TestCase) -> None:
        """Calculates the number of passed, failed, skipped and
        error tests in the test class.

        Args:
            test_case: Test case to be analyzed.
        """
        if test_case.result == "passed":
            self.passed += 1
        if test_case.result == "failed":
            self.failed += 1
            self.failed_test_cases.append(test_case)
        if test_case.result == "skipped":
            self.skipped += 1
            self.skipped_test_cases.append(test_case)
        if test_case.result == "error":
            self.errors += 1
            self.error_test_cases.append(test_case)

        self.execution_time += test_case.execution_time


@dataclass
class TestSuite:
    """Represents a test suite.

    Args:
        name: Name of the test suite.
        tests: Total number of tests.
        errors: Total number of errors.
        failures: Total number of failures.
        skipped: Total number of skipped tests.
        execution_time: Total execution time of the test suite.
        timestamp: Timestamp of the test suite.
        test_classes: List of test classes in the test suite.
        test_cases: List of ungrouped test cases in the test suite.
        passed: Number of passed tests in the test suite.
    """

    name: Optional[str] = field(default=None)
    tests: int = field(default=0)
    errors: int = field(default=0)
    failures: int = field(default=0)
    skipped: int = field(default=0)
    execution_time: float = field(default=0.0)
    timestamp: Optional[str] = field(default=None)
    test_classes: List[TestClass] = field(default_factory=list)
    test_cases: List[TestCase] = field(default_factory=list)
    passed: int = field(init=False, default=0)

    def __post_init__(self) -> None:
        """Calculates the number of passed tests in the test suite."""
        self.passed = self.tests - self.errors - self.failures - self.skipped



"""Copyright (c) 2023, Aydin Abdi.

Command line interface for the qa-analytics-insights package.
"""
import argparse
import sys
from typing import List, Optional  # noqa: F401

from loguru import logger

from qa_analytics_insights import __version__
from qa_analytics_insights.log import (
    default_logging,
    log_execution_time,
    verbose_logging,
)
from qa_analytics_insights.result_analyzer import ResultAnalyzer
from qa_analytics_insights.result_visualizer import ParallelResultVisualizer


class ArgsParser:
    """Class for handling command line arguments."""

    def __init__(self) -> None:
        """Initialize the ArgsParser class."""
        self._parser = None  # type: Optional[argparse.ArgumentParser]
        self._args = None  # type: Optional[argparse.Namespace]

    def add_arguments(self) -> argparse.ArgumentParser:
        """Add command line arguments."""
        parser = argparse.ArgumentParser(
            prog="qa-analytics-insights",
            usage="%(prog)s -f <path/to/files> [options]",
            description="Analyze test reports and generate visualizations.",
            epilog=("Documentation: 'https://qa-analytics-insights.readthedocs.io'."),
        )

        parser.add_argument(
            "-f",
            "--file",
            dest="file_path",
            type=str,
            required=True,
            help="Path to directory or file containing the test results.",
        )
        parser.add_argument(
            "-o",
            "--output",
            dest="output",
            type=str,
            default="test_results_visualization",
            help="Path to the output directory.",
        )
        parser.add_argument(
            "-v",
            "--version",
            action="version",
            version=__version__,
            help="Show the version of the program.",
        )
        parser.add_argument(
            "-vv",
            "--verbose",
            action="store_true",
            dest="verbose",
            help="Enable verbose logging.",
        )
        return parser

    @property
    def parser(self) -> argparse.ArgumentParser:
        """Return the parser object.

        Returns:
            ArgumentParser object.
        """
        if not self._parser:
            self._parser = self.add_arguments()
        return self._parser

    @property
    def args(self) -> argparse.Namespace:
        """Return the parsed arguments.

        Returns:
            Namespace object.
        """
        if not self._args:
            self._args = self.parser.parse_args()
        return self._args

    def help(self) -> None:
        """Print help message."""
        self.parser.print_help()

    def usage(self) -> None:
        """Print usage message."""
        self.parser.print_usage()


class Cli:
    """Class for handling command line interface."""

    def __init__(self) -> None:
        """Initialize the cli class."""
        self.args_parser = ArgsParser()
        self.args = self.args_parser.args

    def run(self, file_path: str, output: str = "test_results_visualization") -> None:
        """Main execution method."""
        test_result_analyzer = ResultAnalyzer(file_path)
        slowest_test_classes = test_result_analyzer.get_slowest_test_classes()
        test_suites = test_result_analyzer.suites
        parallel_test_result = ParallelResultVisualizer(test_suites)
        parallel_test_result.generate_html_plots(output, slowest_test_classes)

    def cli_main(self, args: Optional[List[str]] = None) -> None:
        """Main method for the command line interface.

        Args:
            args: Command line arguments.
        """
        if args is None:
            args = sys.argv[1:]
        if not args:
            self.args_parser.usage()
            return None
        self.args = self.args_parser.parser.parse_args(args)
        if self.args.verbose:
            verbose_logging()
            logger.info("Log level set to DEBUG.")
        else:
            default_logging()
            logger.info("Log level set to INFO.")
        if self.args.file_path is None:
            self.args_parser.help()
            return None
        logger.info("Starting for creating the visualization...")
        logger
        self.run(self.args.file_path, self.args.output)
        logger.info("Visualization created successfully.")


@log_execution_time
def main(args: Optional[List[str]] = None) -> None:
    """Main method for the command line interface.

    Args:
        args: Command line arguments.
    """
    cli = Cli()
    cli.cli_main(args=args)


if __name__ == "__main__":
    """Entry point."""
    main()



"""Unittests for the xml_filter module."""
from pathlib import Path
from queue import Queue

from qa_analytics_insights.xml_filter import XMLFilter


def test_filter_xml() -> None:
    """
    Test the filter_xml method of XMLFilter.

    Args:
        xml_filter_instance (XMLFilter): An instance of XMLFilter.
    """
    path_queue = Queue()
    path_queue.put(Path("tests/data/nosetests_test_result.xml"))
    path_queue.put(Path("tests/data/text.txt"))
    xml_filter = XMLFilter(path_queue)
    xml_queue = xml_filter.filter_xml()
    assert xml_queue.get() == Path("tests/data/nosetests_test_result.xml")
    assert xml_queue.empty()



"""Copyright (c) 2023, Aydin Abdi.

This module processes files in the given path in parallel and
puts the xml files in a queue for further processing.
"""
import threading
from queue import Queue
from typing import List  # noqa: F401

from qa_analytics_insights.data_classes import TestSuite  # noqa: F401
from qa_analytics_insights.patch_fetcher import PathFetcher
from qa_analytics_insights.xml_filter import XMLFilter
from qa_analytics_insights.xml_loader import XMLLoader
from qa_analytics_insights.xml_parser import XMLParser
from qa_analytics_insights.xml_tag_finder import XMLTagFinder


class XMLProcessor:
    """Responsible for processing files into xml queues.

    Args:
        path: Path to the XML files.
    """

    def __init__(self, path: str) -> None:
        """Responsible for processing XML files in the given path.

        Args:
            path: Path to the XML files.
        """
        self.path = path
        self.test_suites = []  # type: List[TestSuite]
        self.lock = threading.Lock()  # type: threading.Lock

    def process_files_in_parallel(self, num_threads: int) -> None:
        """Processes the XML files in the given path in parallel.

        Args:
            num_threads: Number of threads to use for processing.
        """
        file_fetcher = PathFetcher(self.path)
        file_queue = file_fetcher.fetch_paths()
        xml_filter = XMLFilter(file_queue)
        xml_queue = xml_filter.filter_xml()
        threads = []

        for _ in range(num_threads):
            thread = threading.Thread(target=self._process_xml, args=(xml_queue,))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

    def _process_xml(self, xml_queue: Queue[str]) -> None:
        """Processes the XML files in the given path.

        Args:
            xml_queue: Queue of XML files to process.
        """
        while not xml_queue.empty():
            xml_path = xml_queue.get()
            xml_loader = XMLLoader(xml_path)
            xml_tag_finder = XMLTagFinder(xml_loader)
            xml_parser = XMLParser(xml_tag_finder)
            test_suite = xml_parser.parse()

            with self.lock:
                self.test_suites.append(test_suite)



"""Copyright (c) 2023, Aydin Abdi.

Unit tests for XMLTagFinder class.
"""
from xml.etree import ElementTree as ET

import pytest

from qa_analytics_insights.xml_loader import XMLLoader
from qa_analytics_insights.xml_tag_finder import XMLTagFinder


@pytest.fixture
def xml_loader() -> XMLLoader:
    """Returns XMLLoader object."""
    return XMLLoader("tests/data/pytest_test_result.xml")


@pytest.fixture
def xml_tag_finder(xml_loader: XMLLoader) -> XMLTagFinder:
    """Returns XMLTagFinder object."""
    return XMLTagFinder(xml_loader)


def test_test_cases(xml_tag_finder: XMLTagFinder) -> None:
    """Test test_cases property."""
    assert isinstance(xml_tag_finder.test_cases, list)
    assert len(xml_tag_finder.test_cases) == 2
    assert all(isinstance(tag, ET.Element) for tag in xml_tag_finder.test_cases)


def test_suite(xml_tag_finder: XMLTagFinder) -> None:
    """Test suite property."""
    assert isinstance(xml_tag_finder.suite, ET.Element)
    assert xml_tag_finder.suite.tag == "testsuite"


def test_suite_property(xml_tag_finder: XMLTagFinder) -> None:
    """Test suite property."""
    assert xml_tag_finder.suite == xml_tag_finder.root.find("testsuite")



"""Copyright (c) 2023, Aydin Abdi.

This module filters XML files from the given path queue.
"""
from pathlib import Path
from queue import Queue

from loguru import logger


class XMLFilter:
    """Responsible for filtering XML files from the given path queue."""

    def __init__(self, path_queue: Queue[Path]) -> None:
        """Responsible for filtering XML files from the given path queue.

        Args:
            path_queue: Queue of paths to filter XML files from.
        """
        self.path_queue = path_queue
        self.xml_queue = Queue()  # type: Queue[Path]

    def filter_xml(self) -> Queue[Path]:
        """Filters XML files from the given path queue.

        Returns:
            Queue of XML files.
        """
        try:
            while not self.path_queue.empty():
                path = self.path_queue.get()
                if path.suffix == ".xml":
                    self.xml_queue.put(path)
                else:
                    logger.debug(f"Skipped non-XML file: {path}")
        except Exception as unknown_error:  # pragma: no cover
            logger.exception(f"Error filtering XML files: {unknown_error}")
        return self.xml_queue



"""Copyright (c) 2023, Aydin Abdi.

ResultAnalyzer implements the logic for analyzing the test results with
different metrics which can be used to visualize the results.
"""
from typing import List

from qa_analytics_insights.data_classes import TestCase, TestClass, TestSuite
from qa_analytics_insights.xml_processor import XMLProcessor


class ResultAnalyzer:
    """Analyzes the test results with several metrics.

    Example:
        - slowest_test_classes
        - slowest_test_cases
        - failed_test_cases with failure_reason
        - skipped_test_cases with skipped_reason
        - execution times by test class in descending order
    """

    def __init__(self, path: str, num_threads: int = 10) -> None:
        """Initialize the ResultAnalyzer.

        Args:
            path: Path to the test result XML files.
            num_threads: Number of threads to use for parsing the XML files.
        """
        self.path = path
        self.num_threads = num_threads
        self.processor = XMLProcessor(self.path)
        self._suites = []  # type: List[TestSuite]
        self._classes = []  # type: List[TestClass]
        self._test_cases = []  # type: List[TestCase]

    @property
    def suites(self) -> List[TestSuite]:
        """Return the test suites.

        Returns:
            List of TestSuite objects.
        """
        if not self._suites:
            self.process_test_results()
            self._suites = self.processor.test_suites
        return self._suites

    @property
    def classes(self) -> List[TestClass]:
        """Return the test classes.

        Returns:
            List of TestClass objects.
        """
        if not self._classes:
            for test_suite in self.suites:
                self._classes.extend(test_suite.test_classes)
        return self._classes

    @property
    def test_cases(self) -> List[TestCase]:
        """Return the test cases for all test suites.

        Ungrouped test cases are also included.

        Returns:
            List of TestCase objects.
        """
        if not self._test_cases:
            for test_suite in self.suites:
                self._test_cases.extend(test_suite.test_cases)
                for test_class in self.classes:
                    self._test_cases.extend(test_class.test_cases)
        return self._test_cases

    def process_test_results(self) -> None:
        """Process the test result XML files and parse them into TestClass objects.

        Returns:
            List of TestSuites objects.
        """
        self.processor.process_files_in_parallel(self.num_threads)

    def __len__(self) -> int:
        """Return the number of TestClass objects.

        Returns:
            Number of TestClass objects.
        """
        return len(self.test_cases)

    def get_execution_times_by_test_class_in_descending_order(self) -> List[TestClass]:
        """Return the test classes sorted by execution time in descending order.

        Returns:
            List of TestClass objects.
        """
        return sorted(
            self.classes,
            key=lambda test_class: test_class.execution_time,
            reverse=True,
        )

    def get_slowest_test_classes(self, num_test_classes: int = 10) -> List[TestClass]:
        """Return the slowest test classes.

        Args:
            num_test_classes: Number of test classes to return.

        Returns:
            List of TestClass objects.
        """
        return self.get_execution_times_by_test_class_in_descending_order()[
            :num_test_classes
        ]



"""Copyright (c) 2023, Aydin Abdi.

This module is responsible for parsing XML files into TestSuite objects.
"""
from collections import defaultdict
from typing import List  # noqa: F401
from xml.etree import ElementTree as ET  # noqa: F401

from loguru import logger

from qa_analytics_insights.data_classes import TestClass, TestSuite
from qa_analytics_insights.parser import ParserTestCase, ParserTestSuite
from qa_analytics_insights.xml_tag_finder import XMLTagFinder


class XMLParser:
    """Responsible for parsing XML files into TestClass and TestCase objects."""

    def __init__(self, xml_tag_finder: XMLTagFinder) -> None:
        """Responsible for parsing XML files into TestClass and TestCase objects.

        Args:
            xml_tag_finder: XMLTagFinder object.
        """
        self.xml_tag_finder = xml_tag_finder
        self.root = self.xml_tag_finder.root  # type: ET.Element
        self.xml_path = self.xml_tag_finder.xml_loader.xml_path  # type: str
        self.test_case_tags = self.xml_tag_finder.test_cases  # type: List[ET.Element]
        self.suite_tag = self.xml_tag_finder.suite  # type: ET.Element
        self.suite_parser = ParserTestSuite(self.suite_tag)  # type: ParserTestSuite

    def parse(self) -> TestSuite:
        """Parses the XML file into TestClass and TestCase objects.

        This method parses test cases and groups them by their class name.
        If a test case has no class name,it will be added to the
        ungrouped_test_cases list.

        Returns:
            TestSuite object.
        """
        ungrouped_test_cases = []
        classwise_test_cases = defaultdict(list)

        # Creating test cases and grouping them by classname
        for testcase in self.test_case_tags:
            test_case_parser = ParserTestCase(testcase)
            test_case = test_case_parser.parse()
            if not test_case.test_class or test_case.test_class is None:
                logger.debug(f"Test case {test_case.name} has no test class.")
                ungrouped_test_cases.append(test_case)
            else:
                classwise_test_cases[test_case.test_class].append(test_case)

        # Creating TestClass objects from the grouped test cases
        test_classes = [
            TestClass(name=classname, test_cases=test_cases)
            for classname, test_cases in classwise_test_cases.items()
        ]

        # Creating TestSuite object
        test_suite = self.suite_parser.parse()
        test_suite.test_classes = test_classes
        test_suite.test_cases = ungrouped_test_cases

        return test_suite
